/*
  histograms_online.hpp

  Copyright (C) 2016 Taiji Yamada <taiji@aihara.co.jp>
*/
#ifndef _HISTOGRAMS_ONLINE_HPP_
#define _HISTOGRAMS_ONLINE_HPP_
#include "statistics_online.hpp"
#include "histograms.hpp"

template <typename T>
class histogram_online : public statistics_online<T> {
public:
  histogram_online(const size_t n_bins, const T lb, const T ub, const bool logarithmic = false, const bool option_lt_ub = false) : statistics_online<T>(),
    heights(n_bins), lb(lb), ub(ub), logarithmic(logarithmic), option_lt_ub(option_lt_ub), n_prior_points(0), summation(0), prior_points(0) {}
  histogram_online(const size_t n_bins, const size_t n_prior_points, const bool logarithmic = false, const bool option_lt_ub = false) : statistics_online<T>(),
    heights(n_bins), lb(0), ub(0), logarithmic(logarithmic), option_lt_ub(option_lt_ub), n_prior_points(n_prior_points), summation(0)
  {
    if (1 < n_prior_points)
      prior_points = new T[n_prior_points];
  }
  virtual ~histogram_online()
  {
    if (prior_points)
      delete [] prior_points;
    prior_points = 0;
  }
  void add(const T v)
  {
    const size_t count(statistics_online<T>::get_count());
    statistics_online<T>::add(v);
    if (1 < n_prior_points) {
      if (count < n_prior_points) {
	prior_points[count] = v;
	if (count + 1 == n_prior_points) {
	  lb = !logarithmic ?
	    *std::min_element(prior_points, prior_points + n_prior_points) :
	    positive_min_value(prior_points, prior_points + n_prior_points);
	  ub = *std::max_element(prior_points, prior_points + n_prior_points);
	  summation = histogram(heights, prior_points, prior_points + n_prior_points, lb, ub, logarithmic);
	}
      }
      else
	summation += histogram(heights, &v, &v + 1, lb, ub, logarithmic);
    }
    else
      if (!(lb == ub))
	summation += histogram(heights, &v, &v + 1, lb, ub, logarithmic);
  }
  void reset() { statistics_online<T>::reset(); fill(heights.begin(), heights.end(), 0); lb = 0; ub = 0; summation = 0; }

  const std::vector<size_t> &get_heights() const
  {
    return heights;
  }
  size_t get_height(const size_t i) const
  {
    return heights[i];
  }
  T get_lower_bound(const size_t i) const
  {
    return histogram_lower_bound(i, heights.size(), lb, ub, logarithmic, option_lt_ub);
  }
  T get_upper_bound(const size_t i) const
  {
    return histogram_upper_bound(i, heights.size(), lb, ub, logarithmic, option_lt_ub);
  }

  size_t get_minimum_height_index() const
  {
    return std::distance(heights.begin(), min_element(heights.begin(), heights.end()));
  }
  T get_minimum_height() const
  {
    return *min_element(heights.begin(), heights.end());
  }
  size_t get_maximum_height_index() const
  {
    return std::distance(heights.begin(), max_element(heights.begin(), heights.end()));
  }
  T get_maximum_height() const
  {
    return *max_element(heights.begin(), heights.end());
  }

  T get_heights_mean() const
  {
    return histogram_mean<T>(heights, lb, ub, logarithmic);
  }
  T get_heights_variance() const
  {
    return histogram_variance<T>(heights, lb, ub, logarithmic);
  }
  T get_heights_standard_deviation() const
  {
    return histogram_standard_deviation<T>(heights, lb, ub, logarithmic);
  }
  size_t get_heights_summation() const
  {
    return summation;
  }
protected:
  std::vector<size_t> heights;
  T lb, ub;
  bool logarithmic;
  bool option_lt_ub;
  const size_t n_prior_points;
  size_t summation;
  T *prior_points;
};

template <typename T>
class moving_histogram_online : public moving_statistics_online<T> {
public:
  typedef typename moving_statistics_online<T>::internal_value_type internal_value_type;
  moving_histogram_online(const size_t n_bins, const T lb, const T ub, const size_t size = 100, bool logarithmic = false, const bool option_lt_ub = false) : moving_statistics_online<T>(size),
    heights(n_bins), lb(lb), ub(ub), logarithmic(logarithmic), option_lt_ub(option_lt_ub), summation(0) {}
  moving_histogram_online(const size_t n_bins, const size_t size = 100, bool logarithmic = false, const bool option_lt_ub = false) : moving_statistics_online<T>(size),
    heights(n_bins), lb(0), ub(0), logarithmic(logarithmic), option_lt_ub(option_lt_ub), summation(0) {}
  void add(const T v)
  {
    const size_t size(moving_statistics_online<T>::size);
    const size_t index(moving_statistics_online<T>::get_index());
    const T *storage(moving_statistics_online<T>::get_storage());
    const T ov(storage[index]);
    const bool previous_initial(moving_statistics_online<T>::is_initial());
    moving_statistics_online<T>::add(v);
    const bool initial(moving_statistics_online<T>::is_initial());
    if (!initial) {
      if (previous_initial) {
	if (lb == ub) {
	  lb = !logarithmic ?
	    *std::min_element(storage, storage + size) :
	    positive_min_value(storage, storage + size);
	  ub = *std::max_element(storage, storage + size);
	}
	summation = histogram(heights, storage, storage + size, lb, ub, logarithmic);
      }
      else {
	bool update(false);
	if (ov == lb || v < lb) {
	  lb = !logarithmic ?
	    *std::min_element(storage, storage + size) :
	    positive_min_value(storage, storage + size);
	  update = true;
	}
	if (ub == ov || ub < v) {
	  ub = *std::max_element(storage, storage + size);
	  update = true;
	}
	if (update) {
	  fill(heights.begin(), heights.end(), 0);
	  summation = histogram(heights, storage, storage + size, lb, ub, logarithmic);
	}
	else {
	  const internal_value_type
	    sz(!logarithmic ?
	       ub - lb :
	       log(ub) - log(lb));
	  const size_t
	    oid(!logarithmic ?
		(ov - lb)/sz*heights.size() :
		(log(ov) - log(lb))/sz*heights.size()),
	    id(!logarithmic ?
	       (v - lb)/sz*heights.size() :
	       (log(v) - log(lb))/sz*heights.size());
	  if (oid < heights.size() && heights[oid])
	    --heights[oid];
	  if (id < heights.size())
	    ++heights[id];
	}
      }
    }
  }
  void reset() { moving_statistics_online<T>::reset(); fill(heights.begin(), heights.end(), 0); lb = 0; ub = 0; summation = 0; }

  const std::vector<size_t> &get_heights() const
  {
    return heights;
  }
  size_t get_height(const size_t i) const
  {
    return heights[i];
  }
  T get_lower_bound(const size_t i) const
  {
    return histogram_lower_bound(i, heights.size(), lb, ub, logarithmic, option_lt_ub);
  }
  T get_upper_bound(const size_t i) const
  {
    return histogram_upper_bound(i, heights.size(), lb, ub, logarithmic, option_lt_ub);
  }

  size_t get_minimum_height_index() const
  {
    return std::distance(heights.begin(), min_element(heights.begin(), heights.end()));
  }
  T get_minimum_height() const
  {
    return *min_element(heights.begin(), heights.end());
  }
  size_t get_maximum_height_index() const
  {
    return std::distance(heights.begin(), max_element(heights.begin(), heights.end()));
  }
  T get_maximum_height() const
  {
    return *max_element(heights.begin(), heights.end());
  }

  T get_heights_mean() const
  {
    return histogram_mean<T>(heights, lb, ub, logarithmic);
  }
  T get_heights_variance() const
  {
    return histogram_variance<T>(heights, lb, ub, logarithmic);
  }
  T get_heights_standard_deviation() const
  {
    return histogram_standard_deviation<T>(heights, lb, ub, logarithmic);
  }
  size_t get_heights_summation() const
  {
    return summation;
  }
protected:
  std::vector<size_t> heights;
  T lb, ub;
  bool logarithmic;
  bool option_lt_ub;
  size_t summation;
};

#endif
