/*
  multidimensional_histograms_online.hpp

  Copyright (C) 2016 Taiji Yamada <taiji@aihara.co.jp>
*/
#ifndef _MULTIDIMENSIONAL_HISTOGRAMS_ONLINE_HPP_
#define _MULTIDIMENSIONAL_HISTOGRAMS_ONLINE_HPP_
#include "multidimensional_histograms.hpp"
#include "averages_online.hpp"
#include "statistics.hpp"

template <typename T>
class multidimensional_statistics_online {
public:
  typedef typename generic_average_online<T>::internal_value_type internal_value_type;
  multidimensional_statistics_online(const size_t n_dimension) : count(0), n_dimension(n_dimension), positive_minimum_value(n_dimension), negative_maximum_value(n_dimension), minimum_value(n_dimension), maximum_value(n_dimension) {}
  virtual ~multidimensional_statistics_online() {}
  void add(const std::vector<T> &V)
  {
    const std::vector<internal_value_type> v(container_cast<std::vector<internal_value_type> >(V));
    minimum_value = get_count() ? ::min(minimum_value, v) : v;
    maximum_value = get_count() ? ::max(maximum_value, v) : v;
    positive_minimum_value = positive_min(positive_minimum_value, v);
    negative_maximum_value = negative_max(negative_maximum_value, v);
    ++count;
  }
  std::vector<T> get_minimum_value() const { return container_cast<std::vector<T> >(minimum_value); }
  std::vector<T> get_maximum_value() const { return container_cast<std::vector<T> >(maximum_value); }
  std::vector<T> get_positive_minimum_value() const { return container_cast<std::vector<T> >(positive_minimum_value); }
  std::vector<T> get_negative_maximum_value() const { return container_cast<std::vector<T> >(negative_maximum_value); }
  void reset() { count = 0; fill(positive_minimum_value.begin(), positive_minimum_value.end(), 0); fill(negative_maximum_value.begin(), negative_maximum_value.end(), 0); }
  unsigned long long get_count() const { return count; }
protected:
  unsigned long long count;
  const size_t n_dimension;
  std::vector<internal_value_type> positive_minimum_value, negative_maximum_value, minimum_value, maximum_value;
};

template <typename T>
class moving_multidimensional_statistics_online : public multidimensional_statistics_online<T> {
public:
  typedef typename multidimensional_statistics_online<T>::internal_value_type internal_value_type;
  moving_multidimensional_statistics_online(const size_t n_dimension, const size_t n = 100) : multidimensional_statistics_online<T>(n_dimension), size(n), initial(true), index(0)
  {
    storage = new std::vector<T>[size];
  }
  virtual ~moving_multidimensional_statistics_online()
  {
    delete [] storage;
  }
  void add(const std::vector<T> &V)
  {
    const std::vector<internal_value_type> v(container_cast<std::vector<internal_value_type> >(V));
    const std::vector<internal_value_type> ov(container_cast<std::vector<internal_value_type> >(storage[index]));
    if (initial) {
      multidimensional_statistics_online<T>::add(V);
      storage[index++] = V;
      if (index == size) initial = false;
    }
    else {
      storage[index++] = V;
      multidimensional_statistics_online<T>::minimum_value = container_cast<std::vector<internal_value_type> >
	(operator_conditional(operator_equal_to(ov, multidimensional_statistics_online<T>::minimum_value),
			      operator_conditional(operator_less(v, ov),
						   V, min_values(&storage[0], &storage[size])),
			      min_values(&storage[0], &storage[size])));

      multidimensional_statistics_online<T>::maximum_value = container_cast<std::vector<internal_value_type> >
	(operator_conditional(operator_equal_to(ov, multidimensional_statistics_online<T>::maximum_value),
			      operator_conditional(operator_less(ov, v),
						   V, max_values(&storage[0], &storage[size])),
			      max_values(&storage[0], &storage[size])));

      multidimensional_statistics_online<T>::positive_minimum_value = container_cast<std::vector<internal_value_type> >
	(operator_conditional(operator_equal_to(ov, multidimensional_statistics_online<T>::positive_minimum_value),
			      operator_conditional(operator_less(0, v) &&
						   (operator_equal_to(ov, 0) || operator_less(v, ov)),
						   V, positive_min_values(&storage[0], &storage[size])),
			      positive_min_values(&storage[0], &storage[size])));

      multidimensional_statistics_online<T>::negative_maximum_value = container_cast<std::vector<internal_value_type> >
	(operator_conditional(operator_equal_to(ov, multidimensional_statistics_online<T>::negative_maximum_value),
			      operator_conditional(operator_less(v, 0) &&
						   (operator_equal_to(ov, 0) || operator_less(ov, v)),
						   V, negative_max_values(&storage[0], &storage[size])),
			      negative_max_values(&storage[0], &storage[size])));
    }
    if (index == size) index = 0;
  }
  void reset() { multidimensional_statistics_online<T>::reset(); initial = true; index = 0; }
  bool is_initial() const { return initial; }
  size_t get_index() const { return index; }
  const std::vector<T> *get_storage() const { return storage; }
  const size_t size;
private:
  bool initial;
  size_t index;
  std::vector<T> *storage;
};

template <typename T, typename SizeType, typename FlagType>
class multidimensional_histogram_online : public multidimensional_statistics_online<T> {
public:
  multidimensional_histogram_online(const SizeType &n_bins, const std::vector<T> lb, const std::vector<T> ub, const FlagType &logarithmic = false, const bool option_lt_ub = false) : multidimensional_statistics_online<T>(lb.size()),
    n_bins(n_bins), lb(lb), ub(ub), logarithmic(logarithmic), option_lt_ub(option_lt_ub), n_prior_points(0), summation(0), prior_points(0) {}
  multidimensional_histogram_online(const SizeType &n_bins, const size_t n_dimension, const size_t n_prior_points, const FlagType &logarithmic = false, const bool option_lt_ub = false) : multidimensional_statistics_online<T>(n_dimension),
    n_bins(n_bins), lb(n_dimension), ub(n_dimension), logarithmic(logarithmic), option_lt_ub(option_lt_ub), n_prior_points(n_prior_points), summation(0)
  {
    if (1 < n_prior_points)
      prior_points = new std::vector<T>[n_prior_points];
  }
  virtual ~multidimensional_histogram_online()
  {
    if (prior_points)
      delete [] prior_points;
    prior_points = 0;
  }
  void add(const std::vector<T> v)
  {
    const size_t count(multidimensional_statistics_online<T>::get_count());
    multidimensional_statistics_online<T>::add(v);
    if (1 < n_prior_points) {
      if (count < n_prior_points) {
	prior_points[count] = v;
	if (count + 1 == n_prior_points) {
	  lb = operator_conditional(!logarithmic,
				    min_values(prior_points, prior_points + n_prior_points),
				    positive_min_values(prior_points, prior_points + n_prior_points));
	  ub = max_values(prior_points, prior_points + n_prior_points);
	  summation = histogram(heights, prior_points, prior_points + n_prior_points, n_bins, lb, ub, logarithmic);
	}
      }
      else
	summation += histogram(heights, &v, &v + 1, n_bins, lb, ub, logarithmic);
    }
    else
      if (!(lb == ub))
	summation += histogram(heights, &v, &v + 1, n_bins, lb, ub, logarithmic);
  }
  void reset() {
    multidimensional_statistics_online<T>::reset(); heights.clear(); fill(lb.begin(), lb.end(), 0); fill(ub.begin(), ub.end(), 0); summation = 0;
  }

  const std::map<std::vector<size_t>, size_t> &get_heights() const
  {
    return heights;
  }
  size_t get_height(const std::vector<size_t> i) const
  {
    return heights[i];
  }
  std::vector<T> get_lower_bound(const std::vector<size_t> i) const
  {
    return histogram_lower_bound(i, n_bins, lb, ub, logarithmic, option_lt_ub);
  }
  std::vector<T> get_upper_bound(const std::vector<size_t> i) const
  {
    return histogram_upper_bound(i, n_bins, lb, ub, logarithmic, option_lt_ub);
  }

  std::vector<size_t> get_minimum_height_index() const
  {
    return histogram_minimum_height_index(heights);
  }
  T get_minimum_height() const
  {
    return histogram_minimum_height(heights);
  }
  std::vector<size_t> get_maximum_height_index() const
  {
    return histogram_maximum_height_index(heights);
  }
  T get_maximum_height() const
  {
    return histogram_maximum_height(heights);
  }

  std::vector<T> get_heights_mean() const
  {
    return histogram_mean<T, SizeType, FlagType>(heights, n_bins, lb, ub, logarithmic);
  }
  std::vector<T> get_heights_variance() const
  {
    return histogram_variance<T, SizeType, FlagType>(heights, n_bins, lb, ub, logarithmic);
  }
  std::vector<T> get_heights_standard_deviation() const
  {
    return histogram_standard_deviation<T, SizeType, FlagType>(heights, n_bins, lb, ub, logarithmic);
  }
  T get_heights_covariance() const
  {
    return histogram_covariance<T, SizeType, FlagType>(heights, n_bins, lb, ub, logarithmic);
  }
  size_t get_heights_summation() const
  {
    return summation;
  }
protected:
  const SizeType n_bins;
  std::map<std::vector<size_t>, size_t> heights;
  std::vector<T> lb, ub;
  const FlagType &logarithmic;
  bool option_lt_ub;
  const size_t n_prior_points;
  size_t summation;
  std::vector<T> *prior_points;
};

template <typename T, typename SizeType = size_t, typename FlagType = bool>
class moving_multidimensional_histogram_online : public moving_multidimensional_statistics_online<T> {
public:
  typedef typename moving_multidimensional_statistics_online<T>::internal_value_type internal_value_type;
  moving_multidimensional_histogram_online(const SizeType n_bins, const std::vector<T> lb, const std::vector<T> ub, const size_t size = 100, const FlagType &logarithmic = false, const bool option_lt_ub = false) : moving_multidimensional_statistics_online<T>(lb.size(), size),
    n_bins(n_bins), lb(lb), ub(ub), logarithmic(logarithmic), option_lt_ub(option_lt_ub), summation(0) {}
  moving_multidimensional_histogram_online(const SizeType n_bins, const size_t n_dimension, const size_t size = 100, const FlagType &logarithmic = false, const bool option_lt_ub = false) : moving_multidimensional_statistics_online<T>(n_dimension, size),
    n_bins(n_bins), lb(n_dimension), ub(n_dimension), logarithmic(logarithmic), option_lt_ub(option_lt_ub), summation(0) {}
  void add(const std::vector<T> &v)
  {
    const size_t size(moving_multidimensional_statistics_online<T>::size);
    const size_t index(moving_multidimensional_statistics_online<T>::get_index());
    const std::vector<T> *storage(moving_multidimensional_statistics_online<T>::get_storage());
    const std::vector<T> ov(storage[index]);
    const bool previous_initial(moving_multidimensional_statistics_online<T>::is_initial());
    moving_multidimensional_statistics_online<T>::add(v);
    const bool initial(moving_multidimensional_statistics_online<T>::is_initial());
    if (!initial) {
      if (previous_initial) {
	if (/*lb == ub*/exists_operator_equal_to(lb, ub)) {
	  lb = operator_conditional(!logarithmic,
				    min_values(storage, storage + size),
				    positive_min_values(storage, storage + size));
	  ub = max_values(storage, storage + size);
	}
	summation = histogram(heights, storage, storage + size, n_bins, lb, ub, logarithmic);
      }
      else {
	bool update(false);
	if (/*ov == lb*/exists_operator_equal_to(ov, lb) || /*v < lb*/exists_operator_less(v, lb)) {
	  lb = operator_conditional(!logarithmic,
				    min_values(storage, storage + size),
				    positive_min_values(storage, storage + size));
	  update = true;
	}
	if (/*ub == ov*/exists_operator_equal_to(ub, ov) || /*ub < v*/exists_operator_less(ub, v)) {
	  ub = max_values(storage, storage + size);
	  update = true;
	}
	if (update) {
	  heights.clear();
	  summation = histogram(heights, storage, storage + size, n_bins, lb, ub, logarithmic);
	}
	else {
	  const std::vector<internal_value_type>
	    sz(operator_conditional(!logarithmic,
				    container_cast<std::vector<internal_value_type> >(ub - lb),
				    container_cast<std::vector<internal_value_type> >(log(ub) - log(lb))));
	  const std::vector<size_t>
	    oid(operator_conditional(!logarithmic,
				     container_cast<std::vector<size_t> >((ov - lb)/sz*n_bins),
				     container_cast<std::vector<size_t> >((log(ov) - log(lb))/sz*n_bins))),
	    id(operator_conditional(!logarithmic,
				    container_cast<std::vector<size_t> >((v - lb)/sz*n_bins),
				    container_cast<std::vector<size_t> >((log(v) - log(lb))/sz*n_bins)));
	  if (/*oid < n_bins*/forall_operator_less(oid, n_bins) && heights[oid]) {
	    --heights[oid];
	    if (!heights[oid]) heights.erase(oid);
	  }
	  if (/*id < n_bins*/forall_operator_less(id, n_bins))
	    ++heights[id];
	}
      }
    }
  }
  void reset() {
    moving_multidimensional_statistics_online<T>::reset(); heights.clear(); fill(lb.begin(), lb.end(), 0); fill(ub.begin(), ub.end(), 0); summation = 0;
  }

  const std::map<std::vector<size_t>, size_t> &get_heights() const
  {
    return heights;
  }
  size_t get_height(const std::vector<size_t> i) const
  {
    return heights[i];
  }
  std::vector<T> get_lower_bound(const std::vector<size_t> i) const
  {
    return histogram_lower_bound(i, n_bins, lb, ub, logarithmic, option_lt_ub);
  }
  std::vector<T> get_upper_bound(const std::vector<size_t> i) const
  {
    return histogram_upper_bound(i, n_bins, lb, ub, logarithmic, option_lt_ub);
  }

  std::vector<size_t> get_minimum_height_index() const
  {
    return histogram_minimum_height_index(heights);
  }
  T get_minimum_height() const
  {
    return histogram_minimum_height(heights);
  }
  std::vector<size_t> get_maximum_height_index() const
  {
    return histogram_maximum_height_index(heights);
  }
  T get_maximum_height() const
  {
    return histogram_maximum_height(heights);
  }

  std::vector<T> get_heights_mean() const
  {
    return histogram_mean<T, SizeType, FlagType>(heights, n_bins, lb, ub, logarithmic);
  }
  std::vector<T> get_heights_variance() const
  {
    return histogram_variance<T, SizeType, FlagType>(heights, n_bins, lb, ub, logarithmic);
  }
  std::vector<T> get_heights_standard_deviation() const
  {
    return histogram_standard_deviation<T, SizeType, FlagType>(heights, n_bins, lb, ub, logarithmic);
  }
  T get_heights_covariance() const
  {
    return histogram_covariance<T, SizeType, FlagType>(heights, n_bins, lb, ub, logarithmic);
  }
  size_t get_heights_summation() const
  {
    return summation;
  }
protected:
  const SizeType n_bins;
  std::map<std::vector<size_t>, size_t> heights;
  std::vector<T> lb, ub;
  const FlagType &logarithmic;
  bool option_lt_ub;
  size_t summation;
};

#endif
